module E = Errormsg
module I = Input
module L = Logic
module U = Util

open Pretty
open Edag
  
(* This defines the signature that a satisfiability procedure must implement *)

exception Contradiction of Proof.proof   (* a proof of False *)

                                        (* These are the entry points in a 
                                         * decision procedure *)
type entries = 
    {
      canEnable:  Util.symbol list -> bool; 
                                        (** Ask the satproc if it can be used 
                                         * given a proposed list of satprocs *)

      defaultEnabled: bool;              (** If it is enabled by default. If 
                                         * false, it means that you must pass 
                                         * -enable XX to enable it *)

                                         (** Initialize.  *)
      init:       unit -> unit;


      globalAssumeString: unit -> (string * string) option;
                                       (** A name of a parser and a string 
                                        * containing global assumptions *)

      finish:     unit -> unit;

                                        (** Record a snapshot *)
      snapshot:   unit -> unit;
                                        (** Discard all assertions since last 
                                         * not-undone snapshot *)
      undo:       unit -> unit;
                                        (** Assert a literal. Might add new 
                                         * equalities to Satproc.newMessages. 
                                         * Might raise Contradiction, or (if 
                                         * the assertion is a goal) Proved. *)
      assume: Edag.message -> unit;

                                        (** After each assume, this function 
                                         * is called repeatedly until no more 
                                         * messages are produced (in 
                                         * Satproc.newMessages). The argument 
                                         * says if we are about to fail. If 
                                         * in that case you produce some 
                                         * messages, the fixpoint loop is 
                                         * restarted. Thus, to ensure 
                                         * termination stop producing 
                                         * messages in postAssume at some 
                                         * point.  *)
      postAssume: bool -> unit;

                                        (** Accumulate a witness (using 
                                         * {!Edag.accumWitness} for the 
                                         * satisfiability of the goals 
                                         * asserted so far. These can be more 
                                         * specific than the assertions in 
                                         * scope. *)
      witness: unit -> unit;


      debug: bool ref;                  (** For setting a debug flag *)

      ident: Util.symbol;               (** The identifier *)
                
      descr: string;                    (** A short description *)

      options: Options.optionDescr list; (** Command line options *)
    } 


(** Create an empty sat proc with the given name *)
let emptySatproc : entries = 
  { ident = U.registerSymbolName "emptysatproc";
    descr = "missing description";
    options = [];
    debug = ref false;
    canEnable = (fun _ -> true);
    defaultEnabled = true;
    init = (fun _ -> ());
    globalAssumeString = (fun _ -> None);
    finish = (fun _ -> ());
    snapshot = (fun _ -> ());
    undo = (fun _ -> ());
    assume = (fun _ -> ());
    postAssume = (fun _ -> ());
    witness = (fun _ -> ());
  }


(* A specification for a parser *)
type parserSpec = {
    language: string; (* The name of the language. This is used to construct
                       * a command line option --langinput *)
    initParser: string -> Lexing.lexbuf -> unit; (* Initialization *)
    nextFewInputs: Lexing.lexbuf -> Input.input list;

    parserWitness: (Pretty.doc -> unit) -> unit; (** accumulate a witness by 
                                            * calling the provided function *)
  } 


let selectedParser : parserSpec ref = 
  ref { language = "dummy";
        initParser = (fun _ _ -> E.s (E.bug "dummyParser"));
        nextFewInputs = (fun _ -> E.s (E.bug "dummyParser"));
        parserWitness = (fun _ -> E.s (E.bug "dummyParser"));
      }
