#include <stdlib.h>
#include <string.h>
#include "main.h"
#include "input.h"

/* An implementation of values */

/* Up to "nrValues" we have all valid values, but they might be pointing to 
 * unused entries */
NewValue **values;
static int        nrValues = 0; // How many are allocated

int        nrUsedValues = 0;

// This is the size of a value
#define SIZEOF_VALUE ((nrPoints) * sizeof(element) + sizeof(struct NewValue))

// Use the following routine in place of malloc to see how much of space is allocated
static int space = 0, old_space = 0;
inline void* my_malloc(int size)
{
  space = space+size;
  if (old_space + 1000 < space) {
    old_space = space;
    //printf("space = %d Kb, nrValues=%d, nrUsedValues=%d \n",(space/1000), nrValues, nrUsedValues);
  }
  return (malloc(size));
}


NewValue *newValue(void) {
  if(nrUsedValues == nrValues) { // We need to grow the array
    // Decide how many new entries we want. Grow by 50% or at least by 32
    int growBy = nrValues + (nrValues >> 1);
    if(growBy < 32) growBy = 32;

    // We need to realloc the main array
    {
      NewValue **newValues =
        (NewValue**)malloc(sizeof(*newValues) * (nrValues + growBy));
      if(! newValues) _ERROR0("Cannot allocate memory for the new values\n");
      // We need to copy over the old stuff
      memcpy(newValues, values, sizeof(*newValues) * nrValues);
      // Let the old array go
      free(values);
      values = newValues;
    }
    // Now we need to allocate new data for values, in one big block
    {
      char *newValueData = (char*)calloc(growBy, SIZEOF_VALUE);
      int i;
       
      if(!newValueData) _ERROR0("Cannot allocate memory for the new values\n");
      // We need to setup the pointers
      for(i=0;i<growBy;i++) {
        values[nrValues + i] = (NewValue*) &newValueData[i * SIZEOF_VALUE];
      }
    }
    nrValues += growBy;
  }
  _ASSERT(nrUsedValues < nrValues);
  // Now return
  {
    NewValue *res = values[nrUsedValues ++];
    res->used = 1;
    return res;
  }
}

void    releaseValue(NewValue *v) {
  v->used = 0;
}


