module E = Errormsg
module N = Num
open Pretty

(* WARNING: This module is slow since rationals tend to get bigger and bigger 
   with each operation. Therefore, this module should not be preferred for 
   big test-cases *)

module Rational_arithmetic: Arithmetic.arithmetic = struct 
  type elt = N.num


  let docElt () e = dprintf "%s" (N.string_of_num e)
  let eq e1 e2 = (N.eq_num e1 e2)
  let lt e1 e2 = (N.lt_num e1 e2)
  let gt e1 e2 = (N.gt_num e1 e2)
  let leq e1 e2 = (N.le_num e1 e2)
  let geq e1 e2 = (N.ge_num e1 e2)
 
  (* Hashtbl.hash knows how to hash a float but not a num/rational *)
  (* ASSUMPTION: The function N.float_of_num returns Equivalent floats for Equivalent rationals *)
  (* Basis for Assumption: Internet, http://www.bagley.org/~doug/ocaml/Notes/hashtbl.shtml *)
  (* If Assumption Fails: we will loose completeness (not soundness) *)
  (* If assumption is found out to be not true, we should use a different hashing function. 
   * For e.g. normalize the rational and then add numerator and denominator to 
   * get a big integer and then do modulo to get a short int *)  
  let hashElt e = Hashtbl.hash (N.float_of_num e) 

  (* generates a random number between -randBound and +randBound, both exclusive *)
  let randElt () randBound = 
    let doubleRandBound = randBound * 2 in
    N.num_of_int ((Random.int doubleRandBound) - randBound)
  
  let zero = N.num_of_int 0
  let one = N.num_of_int 1
  let int2elt i = N.num_of_int i
  let absolute e = N.abs_num e
  let add e1 e2 = N.add_num e1 e2
  let sub e1 e2 = N.sub_num e1 e2
  let mult e1 e2 = N.mult_num e1 e2
  let div e1 e2 = N.div_num e1 e2
  let join e1 e2 r = N.add_num (N.mult_num e1 r) (N.mult_num e2 (N.sub_num one r))
      
end
    
