// alloc_simple.c
// kernel showing allocation, deallocation and deadness
// "_simple": put (only) needed axiom in the file directly


// --------- verifier preamble ----------
int mem;               // memory for arrays
int meta;              // meta for arrays

extern int tag;        // toplevel address -> type information
extern int objsize;    // toplevel address -> object sizes
extern int objct;      // next address malloc will return


// just identify a pattern term; semantically is the predicate "true"
int PATS __varattribute__((funcsym)) (int pattern ...);

// sel(mem, p) reads from object 'mem' at address 'p'
// SUGAR: mem[p]
int sel __varattribute__((funcsym)) (int mem, int p);

// upd(mem, p, v) yields a new object that is like 'mem',
// except that at location 'p' it has value 'v'
// SUGAR: mem{p := v}
int upd __varattribute__((funcsym)) (int mem, int p, int v);


vc_axiom "sel/upd" {
  // reading from last location written
  forall "sel/upd1" (int obj, int index, int val).
    obj{index := val}[index] == val;
}


int const dead = -1;      // a state value
int const live = 1;       // a state value

// some dumb thing
int metaVar __varattribute__((funcsym)) (int whatever);

//#include "ptr.model"     // sel/upd/ptr
//#include "state.model"   // metaVar


// C functions that are modeled directly
void *malloc(int size);
void free(void *ptr);
void memset(void *ptr, int val, int length);


// -------------------- code to verify ------------------
typedef struct Foo {
  //int use_flat_struct_model;
  int x;      // validity requires: not dead
  int y;      // validity requires: not dead
} Foo;

// invariant: p is dead
Foo * /*owner*/ p;


// global invariant
int global_inv() { return vc_expression(all_of {
  objct > NULL;
  //metaVar(p) == dead;
  p->x == p->x;      // how about '->' in global_inv?
});}


// implied by invariant:
//   pre: p is dead
//   post: p is dead
void func()
  changes(tag, objsize, objct, Foo /*all Foo fields*/, p)
{
  int dummy;

  // p is dead

  p = (Foo * /*owner*/)malloc(sizeof(Foo));

  // p is an owner pointer to a Foo-sized memory area with dead fields
  // temp is dead

  memset(p, 0, sizeof(Foo));

  // p is an owner pointer to a valid Foo (x and y are not dead)

  p->x = 4;      // deref of p is ok because p is valid owner
  dummy = p->x;  // try reading too

  // '->' interpreted inside the VC:
  assert(p->x == 4);

  // and under a 'forall':
  assert(forall(Foo *q) PATS(q->x). p==q ==> q->x==4);
    // the PATS is actually irrelevant because this will be a
    // proof goal, not an assumption, but my vcgen for the moment
    // wants me to put patterns on all foralls ...

  if (p && p->x) {    // should do some path splitting
    p->y = 6;
  }

  // p is an owner pointer to a valid Foo (x and y are not dead)

  free(p);       // ok because p doesn't contain any valid owners,
                 // so its contents are a subtype of "8 deadchars"

  // p is dead
}


/*
  Errors we should be able to catch:
    - deref of p before it's assigned to malloc's return
    - malloc'ing too few bytes
    - use of temp after assigning to p
    - reading x or y from the newly-malloc'd data
    - zero'ing too few bytes: some of Foo still dead
    - zero'ing too many bytes: bounds violation
    - forgetting to free p: doesn't end in 'dead' state
    - freeing p twice
*/
