// clist.c
// linked list of circular linked lists
              

// --------- verifier preamble ----------
#include "ptr.model"        // sel/upd/ptr
#include "state.model"      // metaVar

// C functions that are modeled directly
void *malloc(int size);
void free(void *ptr);


// --------------- code to verify -----------------
// backbone none
struct BNode {
  struct BNode *next;
  struct BNode *prev;      // ghost
  
  struct RNode *ring;
  
  int id;
};

// ring node
struct RNode {
  struct BNode *bnode;     // ghost

  struct RNode *next;
  struct RNode *prev;

  int id;
};

struct BNode *head;


#if 0
int global_inv() { return vc_expression(all_of {
  // head points to first element, if any
  head != NULL ==> all_of {
    tag[head] == BNode_tag;
    head->prev == NULL;
  };

  forall(BNode *b). tag[b]==BNode_tag ==> all_of {
    // injectivity of backbone
    b->next != NULL ==> all_of {
      tag[b->next] == BNode_tag;
      b->next->prev == b;
    };

    // seed into the ring
    tag[b->ring] == RNode_tag;
    b->ring->bnode == b;
  };

  forall(RNode *r). tag[r]==RNode_tag ==> all_of {
    tag[r->next] == RNode_tag;
    tag[r->prev] == RNode_tag;

    // injectivity of ring
    r->next->prev == r;

    // transitivity of 'bnode' label
    //r->next->bnode == r->bnode;
  };
});}
#endif // 0



int global_inv() { return vc_expression(all_of {
  0 < objct;                            // next address isn't NULL
  tag[NULL]==0;                         // zero tag at NULL, i.e. not an object

  // head points to first element, if any
  head != NULL ==> all_of {
    tag[head] == BNode_tag;
    head->prev == NULL;
  };

  forall(BNode *b). tag[b]==BNode_tag ==> all_of {
    // injectivity of backbone
    b->next != NULL ==> all_of {
      tag[b->next] == BNode_tag;
      b->next->prev == b;
    };

    // seed into the ring
    tag[b->ring] == RNode_tag;
    b->ring->bnode == b;
  };

  forall(RNode *r). tag[r]==RNode_tag ==>
    tag[r->next] == RNode_tag;

  forall(RNode *r). tag[r]==RNode_tag ==>
    tag[r->prev] == RNode_tag;

  forall(RNode *r). tag[r]==RNode_tag ==>
    r->next->prev == r;

  forall(RNode *r). tag[r]==RNode_tag ==>
    r->prev->next == r;

  forall(RNode *r). tag[r]==RNode_tag ==>
    r->next->bnode == r->bnode;

  #if 0
  forall(RNode *r). tag[r]==RNode_tag ==> all_of {
    tag[r->next] == RNode_tag;
    tag[r->prev] == RNode_tag;

    // injectivity of ring
    r->next->prev == r;

    // transitivity of 'bnode' label
    //r->next->bnode == r->bnode;
  };
  #endif // 0
});}


// find the ring headed by 'bnodeId', then find the node
// with 'rnodeId' and deallocate it
void deallocRNode(int bnodeId, int rnodeId)
  changes(tag, RNode, BNode)
{
  struct BNode *b;

  for (b=head; b!=NULL; b = b->next) {
    invariant {
      tag[b] == BNode_tag;
    }

    if (b->id == bnodeId) {
      // found the proper ring, now search in it

      struct RNode *r;
      r = b->ring;

      do {
        invariant {
          tag[r] == RNode_tag;
          r->bnode == b;
          metaVar(r) == live;
        }

        if (r->id == rnodeId) {
          // found the proper node, now deallocate it
          
          struct RNode *before = r->prev;
          struct RNode *after = r->next;
          
          if (before == r) {   
            // not really important, but demonstrates the power
            // of the tool to reason about the structure
            assert(after == r);                      
            
            // since 'r' is the only node in the list, don't
            // actually deallocate it
          }
          else {
            assert(after != r);

            // patch around 'r'
            before->next = after;
            after->prev = before;
            if (b->ring == r) {
              b->ring = after;
            }
            
            // deallocate 'r'
            free(r);
          }
          
          return;
        }

        r = r->next;
      } while (r != b->ring);
    }
  }
}























