// gc_copy.c
// 'copy' phase of a copying collector


// --------- verifier preamble ----------
#include "ptr.model"        // sel/upd/ptr
#include "state.model"      // metaVar

// C functions that are modeled directly
void *malloc(int size);

#define GEORGE 
#define EXPANDED

// At the moment, adding typing axioms causes both provers to either
// fail outright, or (if no_objct is specified) take a long time.  So
// this disables them...
verifier_option((no_automatic_typing));


// --------------- code to verify -----------------
enum Color {
  C_WHITE = 0,        // not seen yet
  C_GRAY1 = 1,        // seen but not explored
  C_GRAY2 = 2,        // seen, explored left child
  C_BLACK = 3,        // seen, explored both children
  C_RED = 4,          // element of to-space
};

#define isGrayOrBlack(c) ((c)==C_GRAY1 || (c)==C_GRAY2 || (c)==C_BLACK)

#define notWhite(c) ((c)>=C_GRAY1)

// Somewhat slower with this
//#define notWhite(c) ((c)==C_GRAY1 || (c)==C_GRAY2 || (c)==C_BLACK || (c)==C_RED)

// to be able to turn off things related to 'right' field
#if 1
  #define R(stuff) stuff
#else
  #define R(stuff) /*nothing*/
#endif

// turn off the grid axiom itself
#if 1
  #define GRID(stuff) stuff
#else
  #define GRID(stuff) /*nothing*/
#endif

// turn off assertions that are present just to help me debug
// failed proofs, rather than assertions that should remain
// in the code for programmer assistance, etc.
#if 0
  #define DEBUG_ASSERT(expr) assert(expr) /* user ; */
#else
  #define DEBUG_ASSERT(expr) /*nothing*/ /* user ; */
#endif

// heap modeled as set of pairs
typedef struct Node {
  // programmer-visible fields
  struct Node *left;
R(struct Node *right;)

  // private fields for use by collector
  enum Color color;         // traversal state
  struct Node *fwd;         // forwarding pointer

  // ghost fields for use by verifier
  struct Node *back;        // inverse of 'fwd'
} Node;


// invariant:
//   color is white or gray or black
//   every pointer is valid or NULL
//   if color is white, fwd is NULL
//   if color is gray, fwd->color = red
//   if color is black, fwd->color = red
//      *and* if p->left, p->left->fwd = p->fwd->left
//                        and p->left->color = gray or black
//      and same for right
//   if p->fwd, p->fwd->back = p


int global_inv() { return vc_expression(all_of {
  objct > 0;

  // technical: 'n' is in the used area of the heap
  // I prefer not to use n > 0, because I am not yet sure my unit-case
  // optimization will understand that n != 0
  forall(Node *n) PATS(tag[n]).
     tag[n]==Node_tag ==> n != 0 && n < objct;
  

  // color is white or gray or black or red
  forall(Node *n) PATS(tag[n]). tag[n]==Node_tag ==>
    n->color == C_WHITE ||
    n->color == C_GRAY1 ||
    n->color == C_GRAY2 ||
    n->color == C_BLACK ||
    n->color == C_RED;

  // every pointer is valid or NULL
  forall(Node *n) PATS(n->left).
    tag[n]==Node_tag && n->left!=0  ==> tag[n->left]==Node_tag;

R(
  forall(Node *n) PATS(n->right).                                 
    tag[n]==Node_tag && n->right!=0 ==> tag[n->right]==Node_tag;
 )

  forall(Node *n) PATS(n->fwd).
    (tag[n]==Node_tag && n->fwd!=NULL ==> tag[n->fwd]==Node_tag);

  forall(Node *n) PATS(n->back).
    tag[n]==Node_tag && n->back!=0 ==>tag[n->back]==Node_tag;

  // if color is white or red, fwd is NULL
  forall(Node *n) PATS(n->fwd).
    tag[n]==Node_tag && (n->color==C_WHITE || n->color==C_RED) ==> n->fwd==0;

  // if color is gray or black, fwd is red
  forall(Node *n) PATS(n->fwd).
    tag[n]==Node_tag && isGrayOrBlack(n->color) ==>
      tag[n->fwd]==Node_tag && n->fwd->color==C_RED;     // disequality
  
  // if color is gray2 or black, left is gray or black
  forall(Node *n) PATS(n->left).
    tag[n]==Node_tag && (n->color==C_GRAY2 || n->color==C_BLACK) &&
      n->left!=0 ==> isGrayOrBlack(n->left->color);


R(
  // if color is black, right is gray or black
  forall(Node *n) PATS(n->right).
    tag[n]==Node_tag && n->color==C_BLACK &&
      n->right!=0 ==> isGrayOrBlack(n->right->color);
)

  // if p->fwd, p->fwd->back = p
  forall(Node *n) PATS(n->fwd).
    tag[n]==Node_tag && n->fwd!=0 ==> n->fwd->back == n;                       // injectivity

  // white/gray/black nodes don't reach red nodes
  forall(Node *n) PATS(n->left).
    tag[n]==Node_tag && (n->color==C_WHITE || isGrayOrBlack(n->color)) &&
      n->left!=0  ==> n->left->color  != C_RED;

R(
  forall(Node *n) PATS(n->right).
    tag[n]==Node_tag && (n->color==C_WHITE || isGrayOrBlack(n->color)) &&
    n->right!=0 ==> n->right->color != C_RED;
)

  // left/fwd grid
GRID(
  forall(Node *gridvar) PATS(MPAT(gridvar->left->fwd, gridvar->fwd),
                             MPAT(gridvar->left, gridvar->fwd->left)).
    tag[gridvar]==Node_tag &&
    (gridvar->color==C_GRAY2 || gridvar->color==C_BLACK) &&
    gridvar->left!=0 ==>
      gridvar->left->fwd == gridvar->fwd->left;                                  // grid
)

  // right/fwd grid
GRID(R(
  forall(Node *gridvar) PATS(MPAT(gridvar->right->fwd, gridvar->fwd),
                             MPAT(gridvar->right, gridvar->fwd->right)).
    tag[gridvar]==Node_tag &&
    gridvar->color==C_BLACK &&
    gridvar->right!=0 ==>
      gridvar->right->fwd == gridvar->fwd->right;                                // grid
))

  });}


// just to check that global inv holds
void checkGlobalInv();


void copy(Node *n)
  pre(tag[n]==Node_tag && n->color == C_WHITE)
  post(n->color == C_BLACK &&
       pre(objct) < objct &&

#ifdef GEORGE
       all_of {
         // Tag is preserved
         forall(Node *w) PATS(tag[w], pre(tag[w])).
           (pre(tag)[w]==Node_tag && w != NULL ==> tag[w]==Node_tag);
         
         // color is preserved
         forall(Node *w) PATS(w->color, pre(w->color)).
            (pre(tag)[w]==Node_tag && notWhite(pre(w->color))
                    ==> pre(w->color) == w->color);
         // fwd is preserved
         forall(Node *w) PATS(w->fwd, pre(w->fwd)).
           (pre(tag)[w]==Node_tag && notWhite(pre(w->color))
                    ==> pre(w->fwd) == w->fwd);
         // left is preserved
         forall(Node *w) PATS(w->left, pre(w->left)).
           (pre(tag)[w]==Node_tag && notWhite(pre(w->color))
                    ==>  pre(w->left) == w->left);
         
         // right is preserved
R(         forall(Node *w) PATS(w->right, pre(w->right)).
           (pre(tag)[w]==Node_tag && notWhite(pre(w->color))
                    ==> pre(w->right) == w->right);
 )           
         // back is preserved
         forall(Node *w) PATS(w->back, pre(w->back)).
           (pre(tag)[w]==Node_tag && notWhite(pre(w->color))
                    ==> pre(w->back) == w->back);

       }
#else       
       // have to say that all non-white nodes have not been changed
       forall(Node *w) PATS(pre(tag)[w], tag[w]). 
                        pre(tag)[w]==Node_tag ==> all_of {
                          tag[w]==Node_tag;
                          notWhite(pre(w->color)) ==> all_of {
                            pre(w->left) == w->left;
R(                          pre(w->right) == w->right; )
                            pre(w->color) == w->color;
                            pre(w->fwd) == w->fwd;
                            pre(w->back) == w->back;
                          };
                        }
#endif       
      )
  changes(Node, tag, objct, objsize)
{
  assert(n->fwd == NULL);  

  // add some gratuituous paths
  // int dummy; if(! dummy) {  dummy ++; }
  
  // allocate a node in the to-space
  Node *to = (Node*)malloc(sizeof(Node));

  to->left = NULL;
R(to->right = NULL;)
  to->color = C_RED;
  to->fwd = NULL;
  to->back = n;

  // set up forwarding pointer
  n->fwd = to;
  n->color = C_GRAY1;

  checkGlobalInv();

  // copy children
  if (n->left) {
    assert(n->left->color != C_RED);
    if (n->left->color == C_WHITE) {
      assert(n->fwd == to);
      copy(n->left);
      assert(n->fwd == to);
    }
    checkGlobalInv();
    n->fwd->left = n->left->fwd;
    checkGlobalInv();
  }

  n->color = C_GRAY2;

  DEBUG_ASSERT(
     forall(Node *n). pre(tag)[n]==Node_tag ==> all_of {
                        tag[n]==Node_tag;
                        pre(n->color) >= C_GRAY1 ==> all_of {
                          pre(n->left) == n->left;
R(                        pre(n->right) == n->right; )
                          pre(n->color) == n->color;
                          pre(n->fwd) == n->fwd;
                          pre(n->back) == n->back;
                        };
                      }
  );

R(
  if (n->right) {
    if (n->right->color == C_WHITE) {
      copy(n->right);
    }
    n->fwd->right = n->right->fwd;
  }
)

  n->color = C_BLACK;

  DEBUG_ASSERT(
     forall(Node *n). pre(tag)[n]==Node_tag ==> all_of {
                        tag[n]==Node_tag;
                      }
  );

  DEBUG_ASSERT(
     forall(Node *n). pre(tag)[n]==Node_tag ==> all_of {
                        tag[n]==Node_tag;
                        pre(n->color) >= C_GRAY1 ==> all_of {
                          pre(n->left) == n->left;
R(                        pre(n->right) == n->right; )
                          pre(n->color) == n->color;
                          pre(n->fwd) == n->fwd;
                          pre(n->back) == n->back;
                        };
                      }
  );

  
}






