// privdata.c
// model issues surrounding private_data field of struct file


// ----------------- verifier preamble ----------------
#include "ptr.model"     // sel/upd/ptr
#include "state.model"   // metaVar

// C functions that are modeled directly
void *malloc(int size);
void free(void *ptr);
void memset(void *ptr, int val, int length);


// --------------------- kernel ---------------------
// this declaration comes from headers where we'd like to not be
// forced to add knowledge about every module's use of private_data
struct file {
  int id;                     // invariant: id > 0
  int state;                  // ghost; either 0 or 1
  void *private_data;
};


// --------------------- module ---------------------
typedef struct Scull_dev {
  int y;
} Scull_dev;

#define Scull_dev_inv(addr) (          \
  0 < (addr) < objct &&                \
  tag[addr]==Scull_dev_tag             \
)


// my module id
int const scull_id = 22;


// now in the actual module, we conjoin a new invariant:

// struct file invariant:
//   state==1 && id==2 ==> private_data is a valid pointer to a Scull_dev


int global_inv() { return vc_expression(all_of {
  0 < objct;
  tag[0]==0;

  forall(struct file *addr) PATS(addr->private_data).
    tag[addr]==file_tag && addr->state==1 && addr->id==scull_id ==>
      Scull_dev_inv(addr->private_data);
});}


void noop();


// pre/post: filp is valid
// pre: filp->id == 2  &&   // my module id
//      filp->state == 0    // "closed"
// post: filp->state = 1    // implies something about filp->private_data
void open(struct file *filp)
  pre(0 < filp < objct &&
      tag[filp]==file_tag &&
      filp->id==scull_id &&
      filp->state==0)
  post(filp->state == 1 &&
       forall(void *addr) PATS(tag[addr]). 
         0<addr<pre(objct) ==>
           tag[addr] == pre(tag)[addr])
  changes(tag, objsize, objct, file_state, file_private_data)
{                                
  // debugging problem with failing to say '_pre' on field memory name
  assert(filp->id == scull_id);      // from precondition
  noop();
  if (filp->private_data > 5) { }    // make multiple paths in this fn

  filp->state = 1;          // ghost; "open"
  
  //ERROR(1): filp->state = 0;

  //filp->private_data = (Scull_dev * /*serf*/)&dev;
  filp->private_data = (Scull_dev * /*serf*/)malloc(sizeof(Scull_dev));
}

// pre/post: filp is valid
// pre: filp->id == 2 && filp->state == 1
// post: filp->state == 0
void close(struct file *filp)
  pre(tag[filp]==file_tag &&
      filp->id==scull_id &&
      filp->state==1)
  post(filp->state == 0 &&
       forall(void *addr) PATS(tag[addr]).
         0<addr<pre(objct) ==>
           tag[addr] == pre(tag)[addr])
  changes(Scull_dev_y, file_state)
{
  // cast allowed because of conjoined invariant
  Scull_dev * /*serf*/ p = (Scull_dev * /*serf*/)filp->private_data;

  p->y = 7;     // ok
  //(*p).y = 7;   // also ok; not so much (see star_p.c)

  filp->state = 0;          // ghost    //ERRORIFMISSING(2): required
}


/*
  Errors to catch:
    - use of private_data before it's set, in 'open'
    - failure to set private_data
    - cast private_data to wrong type
*/
