// term.cc
// code for term.h

#include "match.h"          // this module


// ------------------------ MatchRule ---------------------
MatchRule::MatchRule()
{}

MatchRule::~MatchRule()
{}


Term *MatchRule::apply(EDAG &edag, Term *root)
{
  // topmost so far in each branch
  Term *topmost[2];
  topmost[0] = root;
  topmost[1] = root;

  // match each branch against what must already be there
  int branch;            // 0=LHS, 1=RHS
  for (branch=0; branch<2; branch++) {
    for (int i=0; i < match[branch].length(); i++) {
      // is the next symbol represented?
      Term *t = edag.find1(match[branch][i], topmost[branch]);
      if (!t) {
        return NULL;     // rule does not match
      }

      // ok, that becomes the next topmost
      topmost[branch] = t;
    }
  }

  // now, extend using 'get' so nodes will be created if necessary
  for (branch=0; branch<2; branch++) {
    for (int i=0; i < extend[branch].length(); i++) {
      // get the next symbol application
      Term *t = edag.get1(extend[branch][i], topmost[branch]);

      // it becomes the next topmost
      topmost[branch] = t;
    }
  }

  // is the equality already asserted?
  if (topmost[0]->equals(topmost[1])) {
    return NULL;         // rule yielded no new information
  }     
  
  // assert it
  edag.unify(topmost[0], topmost[1]);
  return topmost[0];     // newly unified equivalence class
}


// ------------------------ MatchRuleSet ---------------------
MatchRuleSet::MatchRuleSet()
{}

MatchRuleSet::~MatchRuleSet()
{}


void MatchRuleSet::complete(EDAG &edag)
{
  // iterate until fixpoint
  int changes = 1;
  int iters = 0;
  while (changes) {
    changes = 0;

    // iterate over nodes
    ArrayStack<Term*> nodes;
    edag.getNodeList(nodes);
    for (int i=0; i < nodes.length(); i++) {
      Term *node = nodes[i];

      // iterate over rules
      FOREACH_OBJLIST_NC(MatchRule, rules, iter) {
        // apply
        if (iter.data()->apply(edag, node)) {
          changes++;
        }
      }
    }
    
    iters++;
    if (iters == 1000) {
      xfailure("1000 iterations of rule completion, probable matching loop");
    }
    
    edag.congruenceClosure();
  }
}


// EOF
