// match.h
// module for matching e-dag fragments as a way of implementing axioms

// For now, this is a very simple system, implementing a restricted
// form of axioms such as:
//      +-----+   +--+
//      p->a->b = p->c->d
//      +--+      +-----+

#ifndef MATCH_H
#define MATCH_H

#include "edag.h"       // Symbol
#include "array.h"      // ArrayStack
#include "objlist.h"    // ObjList

// a matching rule
class MatchRule {
public:
  // first, we find a symbol that has the sequence of symbols
  // 'match[0]' applied to it, and also the sequence of symbols
  // 'match[1]'; the 0th element is the innermost (first) symbol
  ArrayStack<Symbol> match[2];

  // then we extend the LHS and RHS by the 'extend' list of symbols,
  // and finally assert the equality of the topmost elements
  ArrayStack<Symbol> extend[2];

public:
  MatchRule();
  ~MatchRule();
  
  // apply the rule, returning non-NULL if
  //   (1) the rule matched at the given root term, and
  //   (2) the asserted equality was new
  // in this case, the return value is the term representing
  // the equivalence class containing the two topmost elements
  Term *apply(EDAG &edag, Term *root);
};


// a set of matching rules, and a procedure for completing
// and EDAG under those rules
class MatchRuleSet {
public:
  // set of rules
  ObjList<MatchRule> rules;
  
public:
  MatchRuleSet();
  ~MatchRuleSet();

  // add a new rule
  void add(MatchRule *r) { rules.append(r); }

  // complete the EDAG under the rules; might not terminate,
  // depending on the rules
  void complete(EDAG &edag);
};


#endif // MATCH_H
