// prv_int.h
// interface to a theorem prover running in a child process

// this is an evolution of the soon-to-be-obsolete prover.h

#ifndef PRV_INT_H
#define PRV_INT_H

#include "str.h"       // string


// when we read some data from the prover, the prover-specific
// object has to say what to do, and does so by returning an
// element of this enumeration
enum ReadAction {
  RA_PROVED,           // prover proved the goal
  RA_FAILED,           // prover failed to prove the goal
  RA_CONTINUE,         // have not seen any decisive output
  RA_ABORT,            // prover has indicated a major problem
};


// object to manage a connection to a prover process
class ProverInterface {
private:     // data
  int toProver, fromProver;    // pipe file descriptors
  int predicateLog;            // file descriptor of log of calls

private:     // disallowed
  ProverInterface(ProverInterface&);
  void operator= (ProverInterface&);

protected:   // funcs
  // return false on EOF
  bool collectMoreInput(stringBuilder &response);
  
  // derived classes must call this upon construction
  void analyzeInit();

protected:   // prover-specific functions
  // return RA_PROVED if all is well
  virtual ReadAction analyzeInitialPrompt(rostring response)=0;

  // return a string for the header of the saved log file
  virtual string getLogfileHeader(rostring fname);

  // look for prover's reply
  virtual ReadAction analyzeGoalResponse(rostring response)=0;

  // do final checks on the proof response; throw exception
  // if there is a problem
  virtual void afterProofCheck(rostring response);

  // given a proof response that provoked RA_FAILED from
  // 'analyzeGoalResponse', search in it for a counterexample;
  // return "" if cannot find it
  virtual string extractCounterexample(rostring response);

public:      // funcs
  ProverInterface(rostring progName);    // start child process
  virtual ~ProverInterface();               // terminate process

  // begin logging predicates to a given file
  void openLog(rostring fname);

  // read a file, and feed all of its contents to the prover; do
  // not expect any output
  void feedFile(rostring fname);

  // feed a string to the prover, again not expecting output
  void feed(rostring str);

  // prove a predicate; returns true if the predicate is proven,
  // false if it is not; 'str' should have a newline at the end;
  // if 'counterexample' is not NULL, and the prover yields a
  // counterexample, put it in that string
  bool prove(rostring str, string *counterexample = NULL);
};

#endif // PRV_INT_H
