// prv_ket.cc
// code for prv_ket.h

#include "prv_ket.h"      // this module
#include "exc.h"          // xbase

#include <string.h>       // strstr


KettleInterface::KettleInterface()
  : ProverInterface("./run-Kettle"),
    respValid("Ready"),
    respInvalid("Continuing after error.*Ready")
{
  analyzeInit();
}

KettleInterface::~KettleInterface()
{}


ReadAction KettleInterface::analyzeInitialPrompt(rostring response)
{
  if (regexpMatch(response, "Starting Kettle.*\n")) {
    return RA_PROVED;
  }
  else {
    return RA_CONTINUE;
  }
}


string KettleInterface::getLogfileHeader(rostring fname)
{
  return stringc
    << "% " << fname << "\n"
    << "% Kettle prover log file\n"
    << "\n";
}


ReadAction KettleInterface::analyzeGoalResponse(rostring response)
{
  if (respInvalid.match(response)) {
    return RA_FAILED;
  }

  if (respValid.match(response)) {
    return RA_PROVED;
  }

  return RA_CONTINUE;
}


string KettleInterface::extractCounterexample(rostring response)
{
  char const *counter = strstr(toCStr(response), "Witness\n");
  char const *inval = strstr(toCStr(response), "*** Continuing");
  xassert(inval);   // just found it a moment ago
  if (counter) {
    //counter += strlen("Counterexample:\n");
    while (*inval != '\n') {
      inval--;
    }
    return substring(counter, inval-counter);
  }
  else {
    return "";    // not found
  }
}


// EOF
