// prv_simp.cc
// code for prv_simp.h

// In essence, this file describes how Simplify's output indicates
// various conditions, like goal proved or not proved.  It does *not*
// encapsulate Simplify's *input* format.  (vml_vcgen.cc does that,
// among many other things.)

#include "prv_simp.h"     // this module
#include "exc.h"          // xbase

#include <string.h>       // strstr


SimplifyInterface::SimplifyInterface()
  : ProverInterface("./run-Simplify"),
    respValid(": Valid."),
    respInvalid(": Invalid."),
    respBad("Bad input"),
    respTriggerless("triggerless quantifier body")
{
  analyzeInit();
}

SimplifyInterface::~SimplifyInterface()
{}


ReadAction SimplifyInterface::analyzeInitialPrompt(rostring response)
{
  if (0==strcmp(response, ">\t")) {
    // turn off prompting
    feed("(PROMPT_OFF)\n");

    return RA_PROVED;
  }
  else {
    return RA_ABORT;
  }
}


string SimplifyInterface::getLogfileHeader(rostring fname)
{
  return stringc
    << "; " << fname << "     -*- lisp -*-\n"
    << "; prover log file\n"
    << "\n";
}


ReadAction SimplifyInterface::analyzeGoalResponse(rostring response)
{
  if (respValid.match(response)) {
    return RA_PROVED;
  }
  
  if (respInvalid.match(response)) {
    return RA_FAILED;
  }
  
  if (respBad.match(response)) {
    return RA_ABORT;
  }
  
  return RA_CONTINUE;
}


void SimplifyInterface::afterProofCheck(rostring response)
{
  // check for the warning about triggerless quantifiers; they are
  // always an indication that something is wrong
  if (respTriggerless.match(response)) {
    char const *processing = strstr(response, "processing the formula");
    char const *after = strstr(processing, ")\n\n");
    if (processing && after) {
      processing += strlen("processing the formula\n\n");
      string formula = substring(processing, (after+1)-processing);
      xbase(stringc << "triggerless: " << formula);
    }
    else {
      xbase("triggerless (couldn't find formula in Simplify output)\n");
    }
  }
}


string SimplifyInterface::extractCounterexample(rostring response)
{
  char const *counter = strstr(response, "Counterexample:");
  char const *inval = strstr(response, ": Invalid.");
  xassert(inval);   // just found it a moment ago
  if (counter) {
    //counter += strlen("Counterexample:\n");
    while (*inval != '\n') {
      inval--;
    }
    return substring(counter, inval-counter);
  }
  else {
    return "";    // not found
  }
}


// EOF
