; reach_f.thy
; reachability predicates and axioms, parameterized by
; the field used for traversal

; requires: memory.thy

; (canReach mem f p q): from p, q is reachable, via field f, in memory state 'mem'
; equivalent to:
;   exists i >= 0, such that access^i(p) = q
;                  where access(p) = (sel mem (ptr p f))
(DEFPRED (canReach mem f p q))

; (canReachWO mem f p q r): from p, q is reachable w/o reading object at r
; equivalent to:
;   exists i >= 0, such that access^i(p) = q,
;                  and forall j < i, access^j(p) != r
;                  where access(p) = (sel mem (ptr p f))
(DEFPRED (canReachWO mem f p q r))

; (noCycle mem f x): no proper path from x to x
(DEFPRED (noCycle mem f x)
  (NOT (canReach mem f (sel mem (ptr x f)) x))   ; f(x) can't reach x
)

(BG_PUSH (AND

;   ; origin of reachability (Nelson A1, restricted to ignore can't-traverse)
;   (FORALL (mem f p q)
;     (IFF
;       (canReach mem f p q)                       ; p reaches q
;       (OR
;         (EQ p q)                                 ; p is q, or
;         (canReach mem f (sel mem (ptr p f)) q)   ; f(p) reaches q
;       )
;     ))

  ; origin of reachability
  (FORALL (mem f p q)
    (IMPLIES
      (EQ (sel mem (ptr p f)) q)
      (canReach mem f p q)
    ))
    
  ; reflexive
  (FORALL (mem f p)
    (canReach mem f p p)
  )

  ; transitive
  (FORALL (mem f p q r)
    (IMPLIES
      (AND
        (canReach mem f p q)               ; p ~> q
        (canReach mem f q r)               ; q ~> r
      )
      (canReach mem f p r)                 ; p ~> r
    ))

  ; if no cycles, x is distinct from everything it can properly reach
  ; (this is easy to prove; is it so provable that I can let Simplify
  ; figure this out on the fly?  for at least one simple example, Simplify
  ; in fact *does* figure it out!)
;   (FORALL (mem f x y)
;     (IMPLIES
;       (AND
;         (noCycle mem f x)
;         (canReach mem f (sel mem (ptr x f)) y)
;       )
;       (NEQ x y)
;     ))

))








