// fnptr.c
// attach pre/post to function pointer type

// ---------------- verifier preamble --------------
// for now, use asm hack to put the includes in
asm("verifier: include \"ptr.vc\";");
asm("verifier: include \"state.vc\";");

// more or less true globals
int mem;

int const live;


// ----------------- code to verify ----------------
int x;


int global_inv() { return vc_expression(all_of {
  1==1;    // "true"
});}


// pre: x > 0
// post: x > pre(x)
typedef void (*FnPtr)()
  pre(x > 0)
  post(x > pre(x))
  changes(x)
;


// pre: x > 0
// post: x > pre(x)
void inc1()
  pre(x > 0)
  post(x > pre(x))
  changes(x)
{
  x++;
}

// pre: x >= 0                weaker than above
// post: x == pre(x)+1        stronger than above
void inc2()
  pre(x >= 0)
  post(x == pre(x)+1)
  changes(x)
{
  x++;
}

#if 0    // conceptually like this
void inc2_wrapper()
  pre(x > 0)
  post(x > pre(x))
  changes(x)
{
  // assume global, x>0
  // prove global, x >= 0
  inc2();
  // assume global, x == pre(x)+1
  // prove global, x > pre(x)
}
#endif // 0

void func()
  changes(x)
{
  // the types match exactly
  FnPtr p = inc1;

  // can now assume that p respects the pre/post
  // attached to FnPtr type
  x = 1;
  (*p)();
  assert(x > 1);

  // Allowing this assignment requires subtyping to take into account
  // the lattice of formula implication.  That might be very hard in
  // general, so formula implication should be decided using a few
  // simple syntactic rules, as opposed to an (unpredictable) appeal
  // to a theorem prover, I think.  However, it's not at all clear
  // that we *need* such flexibility for linux kernel modules.
  p = inc2;
  
  // TODO: actually check the above...

  int y = x;
  (*p)();
  //assert(x == y+1);    // too strong to prove knowing only p's type
  assert(x > y);
}
