/* sxlexer.lex            see license.txt for copyright and terms of use
 * flex description of scanner for s-expressions
 */

/* ----------------------- C definitions ---------------------- */
%{

#include "sxlexer.h"       // Lexer class

// this works around a problem with cygwin & fileno
#define YY_NEVER_INTERACTIVE 1

%}


/* -------------------- flex options ------------------ */
/* no wrapping is needed; setting this means we don't have to link with libfl.a */
%option noyywrap

/* don't use the default-echo rules */
%option nodefault

/* I don't call unput */
%option nounput

/* generate a c++ lexer */
%option c++

/* use the "fast" algorithm with no table compression */
%option full

/* utilize character equivalence classes */
%option ecs

/* the scanner is never interactive */
%option never-interactive

/* and I will define the class (lexer.h) */
%option yyclass="SXLexer"


/* ------------- token definition rules --------------- */
%%
"("                return tok(TOK_LPAREN);
")"                return tok(TOK_RPAREN);

  /* symbol */
[a-zA-Z_0-9]+ {
  return svalTok(TOK_SYMBOL);
}

  /* whitespace */
[ \t\n\f\v\r]+  {
  whitespace();
}

  /* comment */
  /* we don't match the \n because that way this works at EOF */
";".*    {
  whitespace();
}

  /* illegal */
.  {
  err(stringc << "illegal character: `" << yytext[0] << "'");
}

<<EOF>> {
  yyterminate();
}


%%
/**************/
/* extra code */
/**************/
