// sxparse.cc
// code for sxparse.h

#include "sxparse.h"         // this module
#include "sxparse.gr.gen.h"  // SXParser
#include "sxlexer.h"         // SXLexer
#include "glr.h"             // GLR
#include "owner.h"           // Owner
#include "exc.h"             // xbase

TermSeq *parseFile(StringTable &strTable, char const *fname)
{
  // lexer
  SXLexer lexer(strTable, fname);

  // parser context
  SXParse ctx;

  // parse tables (automatically deallocated)
  Owner<ParseTables> tables(ctx.makeTables());

  // parser
  GLR glr(&ctx, tables);

  // parse the input
  SemanticValue treeTop;
  if (!glr.glrParse(lexer, treeTop)) {
    xbase(stringc << "parse error in file " << fname);
  }

  return (TermSeq*)treeTop;
}


// --------------------- test code -----------------------
#ifdef TEST_SXPARSE

#include "srcloc.h"        // SourceLocManager
#include <iostream.h>      // cout

int main(int argc, char *argv[])
{
  if (argc != 2) {
    cout << "usage: " << argv[0] << " input.sx\n";
    return 0;
  }

  SourceLocManager mgr;
  StringTable str;
  Owner<TermSeq> seq; seq = parseFile(str, argv[1]);
  
  FOREACH_ASTLIST(Term, seq->terms, iter) {
    cout << iter.data()->asString() << "\n";
  }

  return 0;
}

#endif // TEST_SXPARSE
