// enum.h
// theory enumerator

#ifndef ENUM_H
#define ENUM_H

#include "objlist.h"           // ObjList
#include "array.h"             // ObjArrayStack

class Rule;                    // rule.h
class ProofTree;               // proof.h
class ProofNode;               // proof.h

// node in the enumeration tree
class EnumNode {
public:      // data
  // predecessor in the enumeration tree; this node is regarded
  // as extending (refining) the proof tree denoted by 'parent';
  // is NULL for a root
  EnumNode *parent;            // (serf)

  // we refine the parent tree by extending it at one particular
  // premise slot; the slot is the one whose rule was added by
  // node 'refinedNode', at premise 'refinedSlot' (0-based)
  // (these are NULL/0 for an enumeration tree root)
  EnumNode *refinedRule;       // (serf)
  int refinedSlot;

  // the rule by which the premise was refined; or for an enumeration
  // tree root, the rule at the root of the proof tree
  Rule *rule;                  // (owner)

public:      // funcs
  EnumNode(EnumNode *parent,
           EnumNode *refined,
           int slot,
           Rule *rule);
  ~EnumNode();
};


// enumeration engine state
class Enumerator {
public:      // data
  // finite set of axioms generating the theory to enumerate
  ObjList<Rule> axioms;
  
  // index for making unique variable names
  int nextRuleIndex;
  
  // queue of nodes of an enumeration tree, which we search
  // in breadth-first order
  ObjArrayStack<EnumNode> queue;

  // index of next element to dequeue; the elements in lower
  // indices have already been processed, but are retained in
  // the array so that they can be deallocated later
  int nextQueueIndex;

  // number of formulas yet to emit; enumerator stops when
  // this becomes nonpositive
  int emitLimit;

private:     // funcs
  ProofTree *buildProofTree(EnumNode *node);
  void enqueueRefinements(EnumNode *enode, ProofNode *n);

public:      // funcs
  Enumerator();
  ~Enumerator();

  // enumerate (print) all the elements of the theory; if
  // 'falseConcl' is true, all emitted consequences are "FALSE"
  void enumerateTheory(int lim, bool falseConcl);
       
  // read/write axioms
  void parseAxioms(char const *fname);
  void printAxioms();
};


#endif // ENUM_H
