// tocnf.h
// convert input sx predicates to CNF (conjunctive normal form)

#ifndef TOCNF_H
#define TOCNF_H

#include "term.h"        // Term
#include "array.h"       // ObjArrayStack, ArrayStack
#include "ptrmap.h"      // PtrMap

              
// stack of assumptions; the entire thing is a conjunction, with each
// stack element being one of:
//   - a single literal (equality or disequality)
//   - a disjunction of literals
//   - a rewrite rule: universally quantified impliciation
//     where premise is conjunction of literals and conclusion
//     is an equality
class CNFStack {
private:     // data
  // some static names for parsing terms
  StringRef str_AND;
  StringRef str_OR;
  StringRef str_IMPLIES;
  StringRef str_NOT;
  StringRef str_FORALL;
  StringRef str_PATS;
  StringRef str_EQ;
  StringRef str_NEQ;
  StringRef str_FALSE;
  StringRef str_TRUE;

  // bidirectional alpha renaming map for variables (StringRef <-> StringRef)
  PtrMap<char const, char const> alpha;      // orig name -> my name
  PtrMap<char const, char const> alphaInv;   // my name -> orig name

  // current set of hypothesis literals for next universal quantifier
  ArrayStack<Term const> hypotheses;

public:      // data
  // table for synthesizing names
  StringTable &str;

  // current set of facts/disjunctions
  ObjArrayStack<Term> facts;

public:      // funcs
  CNFStack(StringTable &str);
  ~CNFStack();

  // convert the given formula into CNF, pushing the conjuncts on the
  // 'facts' stack; if 'sense' is false then we actually push
  // 'not(formula)'; this makes a deep copy of the formula
  void convert(Term const *formula, bool sense);
};


#endif // TOCNF_H
