// vml_build.cc
// code for vml_build.h

#include "vml_build.h"     // this module

Pred *p_and(Pred *p1, Pred *p2)
  { return new P_binary(p1->loc, p1, BP_AND, p2); }
Pred *p_or(Pred *p1, Pred *p2)
  { return new P_binary(p1->loc, p1, BP_OR, p2); }
Pred *p_or3(Pred *p1, Pred *p2, Pred *p3)
  { return p_or(p_or(p1, p2), p3); }
Pred *p_impl(Pred *p1, Pred *p2)
  { return new P_binary(p1->loc, p1, BP_IMPLIES, p2); }
Pred *p_not(Pred *p)
  { return new P_not(p->loc, p); }


Pred *p_compare(VExpr *e1, ComparisonOp op, VExpr *e2)
{                             
  return new P_compare(SL_UNKNOWN, e1, op, e2);
}

Pred *p_compare3(VExpr *e1, ComparisonOp op12, VExpr *e2,
                            ComparisonOp op23, VExpr *e3)
{
  return p_and(p_compare(e1, op12, e2),
               p_compare(e2->clone(), op23, e3));
}

Pred *p_compare4(VExpr *e1, ComparisonOp op12, VExpr *e2,
                            ComparisonOp op23, VExpr *e3,
                            ComparisonOp op34, VExpr *e4)
{
  return p_and(p_compare(e1, op12, e2),
               p_and(p_compare(e2->clone(), op23, e3),
                     p_compare(e3->clone(), op34, e4)));
}


Pred *p_bounds(VExpr *val, VExpr *limit)
{
  return p_compare3(e_lit(0), CO_LESSEQ, val, CO_LESS, limit);
}

Pred *p_equal(VExpr *e1, VExpr *e2)
{
  return p_compare(e1, CO_EQUAL, e2);
}

Pred *p_notEqual(VExpr *e1, VExpr *e2)
{
  return p_compare(e1, CO_NOTEQUAL, e2);
}


Pred *p_forall1(SourceLoc loc, StringRef label, StringRef quantVarName,
                VExpr * /*nullable*/ patspec, Pred *body)
{
  ASTList<VD_var> *vars = new ASTList<VD_var>;
  vars->append(new VD_var(makeId(quantVarName, loc), NULL /*init*/));
  
  return new P_quant(loc, true /*universal*/, label, vars,
                     patspec, body);
}


VExpr *e_neg(VExpr *e)
{
  if (e->isVE_lit()) {
    // special case: make a negative literal (saves space, and might
    // make some things easier)
    VE_lit *el = e->asVE_lit();
    el->n = -(el->n);
    return e;
  }
  else {
    // make a negation node
    return new VE_neg(e);
  }
}

VExpr *e_lit(int i)
{
  return new VE_lit(i);
}

VExpr *e_plus(VExpr *e1, VExpr *e2)
{
  return new VE_arith(e1, AO_PLUS, e2);
}


ASTList<VExpr> *exprlist(VExpr *e1, VExpr *e2, VExpr *e3)
{
  ASTList<VExpr> *ret = new ASTList<VExpr>;
  ret->append(e1);
  ret->append(e2);
  ret->append(e3);
  return ret;
}

ASTList<VExpr> *exprlist(VExpr *e1, VExpr *e2)
{
  ASTList<VExpr> *ret = new ASTList<VExpr>;
  ret->append(e1);
  ret->append(e2);
  return ret;
}

ASTList<VExpr> *exprlist(VExpr *e1)
{
  ASTList<VExpr> *ret = new ASTList<VExpr>;
  ret->append(e1);
  return ret;
}


Identifier *makeId(StringRef s, SourceLoc loc)
{
  return new Identifier(loc, s);
}


VE_apply *apply5(Identifier *func, VExpr *e1, VExpr *e2, VExpr *e3, VExpr *e4, VExpr *e5)
{
  ASTList<VExpr> *args = new ASTList<VExpr>;
  args->append(e1);
  args->append(e2);
  args->append(e3);
  args->append(e4);
  args->append(e5);
  return new VE_apply(func, args);
}

VE_apply *apply4(Identifier *func, VExpr *e1, VExpr *e2, VExpr *e3, VExpr *e4)
{
  ASTList<VExpr> *args = new ASTList<VExpr>;
  args->append(e1);
  args->append(e2);
  args->append(e3);
  args->append(e4);
  return new VE_apply(func, args);
}

VE_apply *apply3(Identifier *func, VExpr *e1, VExpr *e2, VExpr *e3)
{
  ASTList<VExpr> *args = new ASTList<VExpr>;
  args->append(e1);
  args->append(e2);
  args->append(e3);
  return new VE_apply(func, args);
}

VE_apply *apply2(Identifier *func, VExpr *e1, VExpr *e2)
{
  ASTList<VExpr> *args = new ASTList<VExpr>;
  args->append(e1);
  args->append(e2);
  return new VE_apply(func, args);
}

VE_apply *apply1(Identifier *func, VExpr *e1)
{
  ASTList<VExpr> *args = new ASTList<VExpr>;
  args->append(e1);
  return new VE_apply(func, args);
}


// EOF
