// vml_tokens.h
// token codes for VML

#ifndef VML_TOKENS_H
#define VML_TOKENS_H

// there is an array in vml_lexer.cc that contains the strings
// corresponding to these codes

enum VMLToken {
  TOK_EOF = 0,
  
  // multi-spelling
  TOK_INT_LIT,
  TOK_IDENTIFIER,
  TOK_STRING_LIT,

  // NOTE: The quoted strings in comments are significant!  The
  // script that processes this file uses them as the grammar
  // token aliases.

  // keywords
  TOK___PAT,               // "__pat"
  TOK_ANDKEY,              // "AND"
  TOK_AXIOM,               // "axiom"
  TOK_ASSERT,              // "assert"
  TOK_ASSUME,              // "assume"
  TOK_CONST,               // "const"
  TOK_DECLARE_PARTIAL_ORDER,  // "declare_partial_order"
  TOK_ELSE,                // "else"
  TOK_EXISTS,              // "exists"
  TOK_FALSE,               // "false"
  TOK_FORALL,              // "forall"
  TOK_FUNCSYM,             // "funcsym"
  TOK_GOTO,                // "goto"
  TOK_IF,                  // "if"
  TOK_INVARIANT,           // "invariant"
  TOK_ORKEY,               // "OR"
  TOK_POST,                // "post"
  TOK_PRE,                 // "pre"
  TOK_PROCEDURE,           // "procedure"
  TOK_PURE_ASSERT,         // "pure_assert"
  TOK_READS,               // "reads"
  TOK_STANDALONE_INVARIANT,  // "standalone_invariant"
  TOK_TRUE,                // "true"
  TOK_VAR,                 // "var"
  TOK_WHILE,               // "while"
  TOK_WRITES,              // "writes"

  // punctuators, operators
  TOK_LBRACE,              // "{"
  TOK_RBRACE,              // "}"
  TOK_SEMICOLON,           // ";"
  TOK_COLON,               // ":"
  TOK_COLONEQ,             // ":="
  TOK_LPAREN,              // "("
  TOK_RPAREN,              // ")"
  TOK_DOT,                 // "."
  TOK_BANG,                // "!"
  TOK_ANDAND,              // "&&"
  TOK_OROR,                // "||"
  TOK_IMPLIES,             // "==>"
  TOK_IFF,                 // "<==>"
  TOK_EQUAL,               // "=="
  TOK_NOTEQUAL,            // "!="
  TOK_LESS,                // "<"
  TOK_LESSEQ,              // "<="
  TOK_GREATER,             // ">"
  TOK_GREATEREQ,           // ">="
  TOK_PLUS,                // "+"
  TOK_MINUS,               // "-"
  TOK_TIMES,               // "*"
  TOK_LBRACKET,            // "["
  TOK_RBRACKET,            // "]"
  TOK_COMMA,               // ","
  TOK_ARROW,               // "->"
  TOK_SINGLEEQ,            // "="
  TOK_QUESTION,            // "?"

  // misc. for parser
  TOK_HIGH_PREC,
  TOK_LOW_PREC,
  
  NUM_VML_TOKENS,
};


#endif // VML_TOKENS_H
